#include "load_save_png.hpp"
#include <iostream>
#include <glm/glm.hpp>
#include <algorithm>
#include <map>

enum Dir : uint32_t {
	Right = 0,
	UpRight = 1,
	Up = 2,
};

/* probably overkill

struct Edge {
	Edge(int32_t _x, int32_t _y, Dir _dir) : x(_x), y(_y), dir(_dir) { }
	int32_t x,y;
	Dir dir;
};

struct REdge {
	REdge(int32_t _x, int32_t _y, Dir _dir) : x(_x), y(_y), dir(_dir) { }
	int32_t x,y;
	Dir dir;
};

REdge rev(Edge edge) {
	return REdge(edge.x, edge.y, edge.dir);
}
*/

glm::vec2 pixel_to_world(glm::vec2 p) {
	p.x = 70.0f * (p.x / 700.0f - 0.5f);
	p.y = 30.0f * (p.y / 300.0f - 0.5f);
	return p;
}

int main(int argc, char **argv) {
	if (argc != 2) {
		std::cerr << "Usage:\n\tbuild_level <in.png>" << std::endl;
		return 1;
	}
	std::string pngfile = argv[1];
	uint32_t width, height;
	std::vector< uint32_t > data;
	if (!load_png(pngfile, width, height, data, LowerLeftOrigin)) {
		std::cerr << "Failed to load '" << pngfile << "'." << std::endl;
		return 1;
	}

	assert(width == 701);
	assert(height == 301);

	std::string name = pngfile.substr(0, pngfile.size()-4);
	for (uint32_t i = name.size() - 1; i < name.size(); --i) {
		if (name[i] == '/') {
			name = name.substr(i+1);
			break;
		}
	}

	std::cout << "var lev_" << name << " = {\n";

	//-------------------------------------------
	// extract goal positions before filtering
	
	{
		std::multimap< uint32_t, std::pair< glm::vec2, bool > > centers;
		std::vector< uint32_t > mark(data.size(), 0);
		uint32_t fresh_mark = 1;
		for (glm::ivec2 px(0,0); px.y < int32_t(height); ++px.y) {
			for (px.x = 0; px.x < int32_t(width); ++px.x) {
				uint32_t val = data[px.y * width + px.x];
				uint8_t r = val >> 16;
				uint8_t g = val >> 8;
				uint8_t b = val;
				if (r == g && r == b) continue;
				if (mark[px.y * width + px.x] != 0) continue;
				uint32_t m = fresh_mark++;
				//find connected component of pixels with this value.
				std::vector< glm::ivec2 > to_expand;
				std::vector< glm::ivec2 > expanded;
				to_expand.push_back(px);
				mark[px.y * width + px.x] = m;
				std::vector< uint32_t > near(256, 0);
				while (!to_expand.empty()) {
					glm::ivec2 at = to_expand.back();
					to_expand.pop_back();
					expanded.push_back(at);
					assert(mark[at.y * width + at.x] == m);

					static std::vector< glm::ivec2 > nbrs({
						glm::ivec2(-1, 0),
						glm::ivec2( 1, 0),
						glm::ivec2( 0,-1),
						glm::ivec2( 0, 1),
					});
					for (auto d = nbrs.cbegin(); d != nbrs.cend(); ++d) {
						glm::ivec2 n = at + *d;
						if (uint32_t(n.x) < width && uint32_t(n.y) < height) {
							if (mark[n.y * width + n.x] == 0) {
								uint32_t nval = data[n.y * width + n.x];
								if (nval == val) {
									to_expand.push_back(n);
									mark[n.y * width + n.x] = m;
								} else {
									uint8_t r = nval >> 16;
									uint8_t g = nval >> 8;
									uint8_t b = nval;
									if (r == g && r == b) {
										near[r] += 1;
									}
								}
							}
						}
					} //for d in nbrs
				} //while to_expand

				uint32_t mode_count = 0;
				uint32_t mode = 0;
				uint32_t mode_total = 0;
				for (uint32_t i = 0; i < 256; ++i) {
					mode_total += near[i];
					if (near[i] > mode_count) {
						mode = i;
						mode_count = near[i];
					}
				}
				std::cerr << "Mode is " << mode << " with " << mode_count << " of " << mode_total << std::endl;

				//patch:
				for (auto e = expanded.begin(); e != expanded.end(); ++e) {
					data[e->y * width + e->x] = 0xff000000 | (mode << 16) | (mode << 8) | mode;
				}

				//find center:
				glm::vec2 center(0.0f, 0.0f);
				for (auto e = expanded.begin(); e != expanded.end(); ++e) {
					center += glm::vec2(*e);
				}
				center /= expanded.size();
				center = pixel_to_world(center);

				centers.insert(std::make_pair(val, std::make_pair(center, mode == 0)));
			}
		}

		//now that centers are found should probably output them.
		std::cout << "\tgoals:[\n";
		auto c = centers.begin();
		while (c != centers.end()) {
			if (c != centers.begin()) std::cout << ", ";
			uint32_t val = c->first;
			std::vector< std::pair< glm::vec2, bool > > all;
			while (c != centers.end() && c->first == val) {
				all.push_back(c->second);
				++c;
			}
			std::cerr << "Have " << all.size() << " centers for " << (val & 0xff) << std::endl;
			assert(all.size() <= 2);
			if (all.size() == 1) {
				std::cout << "{origin: {x:" << all[0].first.x << ", y:" << all[0].first.y << "},";
				std::cout << "target: {x:" << all[0].first.x << ", y:" << all[0].first.y << "}}";
			} else if (all.size() == 2) {
				assert(all[0].second != all[1].second);
				if (!all[1].second) {
					std::swap(all[0], all[1]);
				}
				std::cout << "{origin: {x:" << all[0].first.x << ", y:" << all[0].first.y << "}, ";
				std::cout << "target: {x:" << all[1].first.x << ", y:" << all[1].first.y << "}}";
			} else {
				assert(0 && "unreachable case");
			}
		}
		std::cout << "],\n";
	}


	//-------------------------------------------
	// now build geometry

	std::vector< uint8_t > grey; grey.reserve(data.size());
	for (auto d = data.begin(); d != data.end(); ++d) {
		grey.push_back(*d & 0xff);
	}

	std::vector< uint32_t > flat_counts(256, 0);
	for (uint32_t y = 1; y + 1 < height; ++y) {
		for (uint32_t x = 1; x + 1 < width; ++x) {
			uint8_t val = grey[y * width + x];
			#define DO(X, Y) \
				if (val != grey[(y+Y) * width + (x+X)]) continue
			DO(-1,-1);
			DO( 0,-1);
			DO( 1,-1);
			DO(-1, 0);
			DO( 1, 0);
			DO(-1, 1);
			DO( 0, 1);
			DO( 1, 1);
			#undef DO
			flat_counts[val] += 1;
		}
	}

	std::vector< uint8_t > flats;

	for (uint32_t i = 0; i < flat_counts.size(); ++i) {
		if (flat_counts[i] > 0) {
			if (flat_counts[i] < 20) {
				std::cerr << " Ignored flat [ " << flat_counts[i] << " ] at " << i << std::endl;
			} else {
				std::cerr << " Flat [ " << flat_counts[i] << " ] at " << i << std::endl;
				flats.push_back(i);
			}
		}
	}

	std::vector< float > levels;
	for (uint32_t i = 0; i + 1 < flats.size(); ++i) {
		levels.push_back((float(flats[i]) + float(flats[i + 1])) * 0.5);
	}
	std::cerr << "Will split at " << levels.size() << " levels." << std::endl;

	std::cout << "\tname:\"" << name << "\",\t\tcontours:[\n";
	for (auto _level = levels.begin(); _level != levels.end(); ++_level) {

		float level = *_level;

		std::cerr << "At " << level << ":" << std::endl;

		level += 0.001f; //"general position"

		std::vector< glm::vec2 > edge_verts(width * height * 3, glm::vec2(-1.0f,-1.0f));

	
		auto get_edge_vert = [&](glm::uvec2 ll, Dir dir) -> glm::vec2 & {
			return edge_verts[(ll.y * width + ll.x) * 3 + dir];
		};

		auto is_vert = [&](uint32_t i) -> bool {
			return edge_verts[i].x >= 0.0f && edge_verts[i].y >= 0.0f;
		};

		auto do_edge = [&](glm::uvec2 ll, Dir dir) {
			uint8_t va = grey[ll.y * width + ll.x];
			uint8_t vb;
			if (dir == Right) {
				vb = grey[ll.y * width + ll.x+1];
			} else if (dir == UpRight) {
				vb = grey[(ll.y+1) * width + ll.x+1];
			} else if (dir == Up) {
				vb = grey[(ll.y+1) * width + ll.x];
			} else {
				assert(0 && "unreachable case");
			}
			if ((va < level) == (vb < level)) {
				get_edge_vert(ll, dir) = glm::vec2(-2.0f,-2.0f);
			} else {
				float amt = (level - va) / (float(vb) - va);
				assert(amt > 0.0f);
				assert(amt < 1.0f);
				glm::vec2 v(ll);
				if (dir == Right || dir == UpRight) {
					v.x += amt;
				}
				if (dir == Up || dir == UpRight) {
					v.y += amt;
				}
				get_edge_vert(ll, dir) = v;
			}
		};

		for (uint32_t y = 0; y < height; ++y) {
			for (uint32_t x = 0; x < width; ++x) {
				if (y + 1 < height) {
					do_edge(glm::uvec2(x,y), Up);
				}
				if (x + 1 < width) {
					do_edge(glm::uvec2(x,y), Right);
				}
				if (y + 1 < height && x + 1 < width) {
					do_edge(glm::uvec2(x,y), UpRight);
				}
			}
		}

		{
			uint32_t verts = 0;
			for (auto v = edge_verts.begin(); v != edge_verts.end(); ++v) {
				if (v->x >= 0.0f && v->y >= 0.0f) {
					++verts;
				}
			}
			std::cerr << "  generated " << verts << " edge verts." << std::endl;
		}

		std::vector< uint32_t > next(width * height * 3, -1U);

		auto connect_tri = [&](uint32_t ab_idx, uint32_t bc_idx, uint32_t ca_idx, bool a_in, bool b_in, bool c_in) {
			if (a_in == b_in && a_in == c_in) return;
			if (a_in == b_in) {
				assert(is_vert(bc_idx));
				assert(is_vert(ca_idx));
				if (c_in) next[ca_idx] = bc_idx;
				else next[bc_idx] = ca_idx;
			} else if (b_in == c_in) {
				assert(is_vert(ab_idx));
				assert(is_vert(ca_idx));
				if (a_in) next[ab_idx] = ca_idx;
				else next[ca_idx] = ab_idx;
			} else if (c_in == a_in) {
				assert(is_vert(ab_idx));
				assert(is_vert(bc_idx));
				if (b_in) next[bc_idx] = ab_idx;
				else next[ab_idx] = bc_idx;
			} else {
				assert(0 && "unreachable case");
			}
		};

		for (uint32_t y = 0; y + 1 < height; ++y) {
			for (uint32_t x = 0; x + 1 < width; ++x) {
				{ //upper triangle:
					uint32_t ab_idx = (y * width + x) * 3 + UpRight;
					uint32_t bc_idx = ((y+1) * width + x) * 3 + Right;
					uint32_t ca_idx = (y * width + x) * 3 + Up;
					bool a_in = grey[y * width + x] > level;
					bool b_in = grey[(y+1) * width + (x+1)] > level;
					bool c_in = grey[(y+1) * width + x] > level;
					connect_tri(ab_idx, bc_idx, ca_idx, a_in, b_in, c_in);
				}
				{ //lower triangle:
					uint32_t ab_idx = (y * width + x) * 3 + Right;
					uint32_t bc_idx = (y * width + x+1) * 3 + Up;
					uint32_t ca_idx = (y * width + x) * 3 + UpRight;
					bool a_in = grey[y * width + x] > level;
					bool b_in = grey[y * width + (x+1)] > level;
					bool c_in = grey[(y+1) * width + (x+1)] > level;
					connect_tri(ab_idx, bc_idx, ca_idx, a_in, b_in, c_in);
				}
			}
		}

		std::vector< uint32_t > prev(width * height * 3, -1U);
		for (uint32_t from = 0; from < next.size(); ++from) {
			uint32_t to = next[from];
			if (to != -1U) {
				assert(prev[to] == -1U);
				prev[to] = from;
			}
		}


		std::vector< bool > used(edge_verts.size(), false);


		//chains touch some edge, loops don't:
		std::vector< std::vector< glm::vec2 > > chains;
		std::vector< std::vector< glm::vec2 > > loops;

		for (uint32_t seed = 0; seed < edge_verts.size(); ++seed) {
			if (used[seed]) continue;
			if (!is_vert(seed)) continue;

			uint32_t at = seed;
			while (prev[at] != -1U) {
				at = prev[at];
				if (at == seed) break;
			}

			uint32_t start = at;
			bool closed = false;
			std::vector< glm::vec2 > chain;
			while (at != -1U) {
				assert(at < edge_verts.size());
				assert(!used[at]);
				used[at] = true;
				assert(is_vert(at));
				chain.push_back(edge_verts[at]);
				at = next[at];
				if (at == start) {
					closed = true;
					break;
				}
			}
			{ //resample the chain with uniform steps:
				std::vector< glm::vec2 > resampled;
				const float Step = 3.5f;
				float along = 0.0f;
				resampled.push_back(chain[0]);
				for (auto c = chain.begin(); c + 1 != chain.end(); ++c) {
					glm::vec2 from = *c;
					glm::vec2 to = *(c + 1);
					float len = glm::length(from - to);
					glm::vec2 vec = (to - from) / len;
					float remain = len;
					while (remain > 0.0f) {
						if (remain >= (Step - along)) {
							remain -= (Step - along);
							along = 0.0f;
							resampled.emplace_back(to - vec * (remain / len));
						} else {
							along += remain;
							remain = 0.0f;
						}
					}
				}
				if (glm::length(resampled.back() - chain.back()) > 0.5f * Step) {
					resampled.push_back(chain.back());
				} else {
					resampled.back() = chain.back();
				}
				chain = resampled;
			}
			if (closed) {
				std::cerr << " Found a loop of size " << chain.size() << "." << std::endl;
				loops.push_back(chain);
			} else {
				std::cerr << " Found a chain of size " << chain.size() << "." << std::endl;
				chains.push_back(chain);
			}
		}

		//make chains into loops:

		auto get_val = [&](glm::vec2 v) -> std::pair< uint32_t, float > {
			bool lr = (v.x == 0.0 || v.x == width - 1);
			bool du = (v.y == 0.0 || v.y == height - 1);
			assert(lr || du);
			assert(!(lr && du));
			if (v.y == 0.0f) return std::make_pair( 0, v.x );
			if (v.x == width - 1) return std::make_pair( 1, v.y );
			if (v.y == height - 1) return std::make_pair( 2, float(width) - 1 - v.x );
			if (v.x == 0.0f) return std::make_pair( 3, float(height) - 1 - v.y );
			std::cerr << "Why is this not on an edge: " << v.x << ", " << v.y << std::endl;
			assert(0 && "unreachable, one hopes");
		};

		std::map< std::pair< uint32_t, float >, uint32_t > pts;

		pts.insert(std::make_pair(std::make_pair(0, 0.0f), -1U));
		pts.insert(std::make_pair(std::make_pair(1, 0.0f), -1U));
		pts.insert(std::make_pair(std::make_pair(2, 0.0f), -1U));
		pts.insert(std::make_pair(std::make_pair(3, 0.0f), -1U));

		const uint32_t EndIdx = chains.size();
		for (auto c = chains.begin(); c != chains.end(); ++c) {
			pts.insert(std::make_pair(get_val((*c)[0]), c - chains.begin()));
			pts.insert(std::make_pair(get_val(c->back()), EndIdx));
		}

		std::vector< bool > used_chains(chains.size(), false);
		for (auto c = chains.begin(); c != chains.end(); ++c) {
			if (used_chains[c - chains.begin()]) continue;
			auto start = pts.find(get_val((*c)[0]));
			assert(start != pts.end());
			assert(start->second == c - chains.begin());

			std::vector< glm::vec2 > loop;
			auto at = start;
			while (1) {
				if (at->second == -1U) {
					if (at->first.first == 0) {
						loop.emplace_back(at->first.second, 0.0f);
					} else if (at->first.first == 1) {
						loop.emplace_back(width - 1, at->first.second);
					} else if (at->first.first == 2) {
						loop.emplace_back(float(width) - 1 - at->first.second, height - 1);
					} else if (at->first.first == 3) {
						loop.emplace_back(0.0f, float(height) - 1 - at->first.second);
					}
				} else {
					assert(at->second != EndIdx);
					assert(at->second < chains.size());
					assert(!used_chains[at->second]);
					used_chains[at->second] = true;
					auto c = chains.begin() + at->second;
					loop.insert(loop.end(), c->begin(), c->end());
					at = pts.find(get_val(c->back()));
					assert(at != pts.end());
				}
				++at;
				if (at == pts.end()) at = pts.begin();
				if (at == start) break;
			}

			loops.push_back(loop);
		}


		if (loops.empty()) std::cerr << "WARNING: found no loops." << std::endl;

		if (_level != levels.begin()) {
			std::cout << ",\n";
		}
		std::cout << "\t{ idx:" << (_level - levels.begin()) << ",\n";
		std::cout << "\tloops: [\n";
		for (auto loop = loops.begin(); loop != loops.end(); ++loop) {
			//add chain (now closed, one hopes) to the list:
			if (loop != loops.begin()) {
				std::cout << ",\n";
			}
			std::cout << "\t\t[ ";
			for (auto v = loop->begin(); v != loop->end(); ++v) {
				if (v != loop->begin()) {
					if ((v - loop->begin()) % 8 == 0) {
						std::cout << ",\n\t\t";
					} else {
						std::cout << ", ";
					}
				}
				std::cout << (v->x - 0.5f * (width - 1)) / 300.0f * 30.0f << ","
				          << (v->y - 0.5f * (height - 1)) / 300.0f * 30.0f;
			}
			std::cout << " ]";
		} //for loop

		std::cout << "\n\t] }";
	} // for level
	std::cout << "\n] };" << std::endl;

	return 0;
}
