/*
 * (c) 2014 Jim McCann / TCHOW llc.
 *  Feel free to learn from this code; please don't steal it.
 */

var ProjectileRadius = 0.5;
var MaxStretch = 4.5;
var MinStretch = 1.0;
var StretchK = 300.0;
var Gravity = 40.0;
var Tick = 1.0 / 60.0;
var BPM = 120.0;
var GoalRad = 0.7;
var BgVolume = 0.5;

var chimeOrder = [];

var sounds = {
	activates:[],
	background:null,
	quarterChimes: [],
	halfChimes: [],
	wholeChimes: [],
	explodes: [],
	pull: null,
	release: null
};

var solidProgram = null;
var textureProgram = null;
var levelbgProgram = null;
var particleProgram = null;
var surfaceProgram = null;

var noiseTexture = null;
var surfaceTexture = null;
var titleTexture = null;
var wellDoneTexture = null;

var ModeNone = "N";
var ModeSelect = "S";
var ModePlay = "P";
var ModeToSelect = ">S";

var mode = ModeNone;

var GoalInactive = ".";
var GoalSpinning = "c";
var GoalExploding = "e";
var GoalActive = "a";

var surfWave = 0.0;

var chimePattern = []; //used by ModeToSelect

var camera = {
	at: { x: 0, y: 0 },
	up: { x: 0, y: 15 },
	//will be set in setup() -- right: {x: 35, y: 0 }
};

function getCameraMatrix() {
	var lenSq = (camera.up.x * camera.up.x + camera.up.y * camera.up.y);
	var invUp = {
		x: camera.up.x / lenSq,
		y: camera.up.y / lenSq
	};

	lenSq = (camera.right.x * camera.right.x + camera.right.y * camera.right.y);

	var invRight = {
		x: camera.right.x / lenSq,
		y: camera.right.y / lenSq
	};
	var Matrix = [
		invRight.x, invUp.x, 0.0, 0.0,
		invRight.y, invUp.y, 0.0, 0.0,
		0.0, 0.0, 1.0, 0.0,
		invRight.x * -camera.at.x + invRight.y * -camera.at.y,
			invUp.x * -camera.at.x + invUp.y * -camera.at.y,
			0.0, 1.0
	];
	return Matrix;
}
function getScaleTranslateMatrix(s, x,y,z) {
	var Matrix = [
		s, 0.0, 0.0, 0.0,
		0.0, s, 0.0, 0.0,
		0.0, 0.0, s, 0.0,
		x, y, z, 1.0
	];
	return Matrix;
}
function getRotateMatrix(x_dir, y_dir, z_dir) {
	var Matrix = [
		x_dir.x, x_dir.y, x_dir.z, 0.0,
		y_dir.x, y_dir.y, y_dir.z, 0.0,
		z_dir.x, z_dir.y, z_dir.z, 0.0,
		0.0, 0.0, 0.0, 1.0,
	];
	return Matrix;
}

function multMatrix(a,b) {
	var Matrix = [
		0.0, 0.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 0.0
	];
	if (a.length != 16) throw "Bad size a";
	if (b.length != 16) throw "Bad size b";
	for (var r = 0; r < 4; ++r) {
		for (var c = 0; c < 4; ++c) {
			for (var i = 0; i < 4; ++i) {
				Matrix[c * 4 + r] +=
					a[i * 4 + r] * b[c * 4 + i];
			}
		}
	}
	return Matrix;
}

function dot2(a,b) {
	return a.x * b.x + a.y * b.y;
}
function dot3(a,b) {
	return a.x * b.x + a.y * b.y + a.z * b.z;
}
function normalize2(vec) {
	var len = Math.sqrt(dot2(vec,vec));
	return {x: vec.x / len, y: vec.y / len};
}
function normalize3(vec) {
	var len = Math.sqrt(dot3(vec,vec));
	return {x: vec.x / len, y: vec.y / len, z: vec.z / len};
}
function cross3(a,b) {
	return {
		x: a.y * b.z - a.z * b.y,
		y: a.z * b.x - a.x * b.z,
		z: a.x * b.y - a.y * b.x
	};
}


function setCameraUniforms() {
	gl.uniformMatrix4fv(gl.getUniformLocation(gl.getParameter(gl.CURRENT_PROGRAM), "Matrix"), false, getCameraMatrix());
}

function setCollisionCameraUniforms() {
	var temp = camera;
	camera = {
		at: {x:0, y:0},
		up: {x:0, y:15},
		right: {x:35, y:0},
	};
	setCameraUniforms();
	camera = temp;
}

function playSound(sound, volume) {
	if (sound == null) return null; //not yet loaded
	if (sound === sounds.quarterChimes || sound === sounds.halfChimes || sound === sounds.wholeChimes) {
		var i = chimeOrder.shift();
		//console.log("Chime: " + i);
		sound = sound[i];
		chimeOrder.push(i);
		if (sound === undefined) return;
	} else if (sound instanceof Array) {
		var avail = [];
		sound.forEach(function(s){
			if (s.paused) {
				avail.push(s);
			}
		});
		if (avail.length == 0) {
			console.log("all sounds busy");
			return null; //all sounds busy
		}
		sound = avail[Math.random() * avail.length | 0];
	}
	if (!sound.paused) { //sound is busy
		return null;
	}
	if (volume === undefined) {
		volume = 1.0;
	}
	if (muted) volume = 0.0;
	sound.volume = volume;
	sound.currentTime = 0.0;
	sound.play();
	return sound;
}

var mouse = {
	down:false,
	x:0,y:-20,
	prevx:[], prevy:[]
};

var followers = [{x: 0, y: 0, rad: 0.2}];

var projectile = undefined;
var nearGoal = undefined;

var chimes = [];
var chimePairs = [];
var chimeWait = 0.0;

function mousedown() {
	if (!mouse.down) {
		mouse.down = true;
		if (mode === ModeSelect) {
			for (var i = 0; i < level.goals.length; ++i) {
				if (level.goals[i] == nearGoal) {
					set_level(levels[i]);
					playSound(sounds.explodes);
					break;
				}
			}
		} else if (mode === ModePlay) {
			if (projectile === undefined) {
				if (nearGoal !== undefined) {
					projectile = {
						x: mouse.x, y: mouse.y,
						vx: 0.0, vy: 0.0,
						launched: 0.0,
						prevx: [], prevy: [],
					};
					if (sounds.pull !== null) {
						sounds.pull.pause();
						sounds.pull.currentTime = 0.0;
						if (muted) {
							sounds.pull.volume = 0.0;
						} else {
							sounds.pull.volume = 1.0;
						}
						sounds.pull.play();
					}
					if (sounds.release !== null) {
						sounds.release.pause();
						sounds.release.currentTime = 0.0;
						sounds.release.volume = 0.0;
						sounds.release.play();
					}
				}
			};
		}
	}
}

function mouseup() {
	if (mouse.down) {
		mouse.down = false;
		if (mode === ModeSelect) {
			//(do nothing)
		} else if (mode === ModePlay) {
			if (projectile !== undefined) {
				if (projectile.launched == 0.0) {
					projectile.launched = 0.001;
					var to = {
						x: nearGoal.at.x - projectile.x,
						y: nearGoal.at.y - projectile.y,
					};
					projectile.launchTo = to;
					var launchA = Math.sqrt(to.x * to.x + to.y * to.y);
					projectile.launchF = Math.sqrt(StretchK / launchA);
					projectile.vx = to.x * projectile.launchF;
					projectile.vy = to.y * projectile.launchF;
					if (launchA < MinStretch) {
						projectile = undefined;
					} else {
						level.launches += 1;
					}
					if (sounds.pull !== null) {
						sounds.pull.volume = 0.0;
					}
					if (sounds.release !== null) {
						if (muted) {
							sounds.release.volume = 0.0;
						} else {
							sounds.release.volume = 1.0;
						}
					}
				}
			}
		}
	}
}

function mousemove(pos) {
	var delta = {x:pos.x - mouse.x, y:pos.y - mouse.y};
	var len = Math.sqrt(delta.x * delta.x + delta.y * delta.y);
	var newLen = Math.log(1 + len);
	delta.x *= newLen / len;
	delta.y *= newLen / len;

	mouse.x = pos.x;
	mouse.y = pos.y;

	followers.forEach(function(f){
		f.x += delta.x * f.t;
		f.y += delta.y * f.t;
	});


	if (projectile !== undefined) {
		if (projectile.launched == 0.0) {
			var to = {
				x: mouse.x - nearGoal.at.x,
				y: mouse.y - nearGoal.at.y
			};
			var len = Math.sqrt(to.x * to.x + to.y * to.y);
			if (len > MaxStretch) {
				to.x *= MaxStretch / len;
				to.y *= MaxStretch / len;
			}
			projectile.x = nearGoal.at.x + to.x;
			projectile.y = nearGoal.at.y + to.y;

		} else {
			//not much to do, I guess...
		}
	} else {
		update_nearGoal();
	}
}

function update_nearGoal() {
	if (projectile === undefined) {
		nearGoal = undefined;
		var nearDisSq = 2.0 * 2.0;
		level.goals.forEach(function(goal){
			var disSq = (goal.at.x - mouse.x) * (goal.at.x - mouse.x) + (goal.at.y - mouse.y) * (goal.at.y - mouse.y);
			if (disSq < nearDisSq) {
				nearDisSq = disSq;
				nearGoal = goal;
			}
		});
		if (nearGoal !== undefined) {
			if (mode === ModeSelect) {
				if (nearGoal.state !== GoalActive) {
					nearGoal = undefined;
				}
			} else if (mode === ModePlay) {
				if (nearGoal.state !== GoalActive) {
					nearGoal = undefined;
				}
			} else {
				nearGoal = undefined;
			}
		}
	}
}

var collisionFb = null;
var collisionTexture = null;

var level = null;

var scores = [];

function set_select() {
	//TODO: transition stuff.
	mode = ModeSelect;

	//-----------------------------------------------
	level = {goals:[]};

	var first_inactive = true;
	for (var i = 0; i < levels.length; ++i) {
		level.goals.push({
			origin:{x: (i - 0.5 * (levels.length-1)) * 3.0, y:-6.0},
		});
		init_goal(level.goals[level.goals.length-1]);
		if (scores[i] !== undefined) {
			level.goals[i].state = GoalActive;
		} else if (first_inactive) {
			level.goals[i].state = GoalActive;
			level.goals[i].glow = {rad:1.0, alpha:0.1, cycle:Math.random()};
			first_inactive = false;
		}
	}

	chimes = [];
	chimePairs = [];

	if (collisionFb) {
		gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
		gl.disable(gl.BLEND);
		gl.clearColor(0.0, 0.0, 0.0, 0.0);
		gl.stencilMask(0xff);
		gl.clear(gl.COLOR_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);
		gl.bindFramebuffer(gl.FRAMEBUFFER, null);
	}

}

function init_goal(goal) {
	delete goal.glow;
	delete goal.explodeAttribs;
	delete goal.tween;
	goal.at = {x:goal.origin.x, y:goal.origin.y};
	goal.state = GoalInactive;
	goal.lines = make_goal_lines();
	goal.x_dir = normalize3({x:Math.random(), y:Math.random(), z:Math.random()});
	goal.y_dir = {x:Math.random(), y:Math.random(), z:Math.random()};
	var f = dot3(goal.x_dir, goal.y_dir);
	goal.y_dir.x -= f * goal.x_dir.x;
	goal.y_dir.y -= f * goal.x_dir.y;
	goal.y_dir.z -= f * goal.x_dir.z;
	goal.y_dir = normalize3(goal.y_dir);
	goal.z_dir = cross3(goal.x_dir, goal.y_dir);
}

function make_goal_lines() {
	var Count = 4;
	var verts = [];
	var edges = [];

	var sin60 = Math.sin(60.0 / 180.0 * Math.PI);
	var cos60 = Math.cos(60.0 / 180.0 * Math.PI);
	verts = [
		cos60, sin60, 0.0,
		cos60,-sin60, 0.0,
		-cos60, 0.0, sin60,
		-cos60, 0.0,-sin60,
	];
	edges = [
		0,3, 0,6, 0,9,
		3,6, 3,9,
		6,9
	];

	/*
	for (var i = 0; i < verts.length; ++i) {
		verts[i] = Math.random() - 0.5;
	}
	for (var i = 0; i + 2 < verts.length; i += 3) {
		var x = verts[i];
		var y = verts[i+1];
		var z = verts[i+2];
		var len = Math.sqrt(x*x + y*y + z*z);
		if (len > 0) {
			x /= len;
			y /= len;
			z /= len;
		} else {
			x = 1.0;
			y = 0.0;
			z = 0.0;
		}
		verts[i] = x;
		verts[i+1] = y;
		verts[i+2] = z;
	}
	var Thresh = 1.0;
	for (var i = 0; i + 2 < verts.length; i += 3) {
		for (var j = i; j + 2 < verts.length; j += 3) {
			var len = (verts[i] - verts[j]) * (verts[i] - verts[j])
			        + (verts[i+1] - verts[j+1]) * (verts[i+1] - verts[j+1])
			        + (verts[i+2] - verts[j+2]) * (verts[i+2] - verts[j+2]);
			if (len < Thresh * Thresh) {
				edges.push(i,j);
			}
		}
	}
	//TODO: smooth and iterate?
	*/

	var lines = new Float32Array(edges.length * 6);
	for (var i = 0; i + 1 < edges.length; i += 2) {
		lines[i*3+0] = verts[edges[i]];
		lines[i*3+1] = verts[edges[i]+1];
		lines[i*3+2] = verts[edges[i]+2];
		lines[i*3+3] = verts[edges[i+1]];
		lines[i*3+4] = verts[edges[i+1]+1];
		lines[i*3+5] = verts[edges[i+1]+2];
	}
	return lines;
	/*
	var lengths = new Float32Array(Count * Count);
	for (var i = 0; i + 1 < verts.length; i += 2) {
		for (var j = i; j + 1 < verts.length; j += 2) {
			var len = (verts[i] - verts[j]) * (verts[i] - verts[j]) + (verts[i+1] - verts[j+1]) * (verts[i+1] - verts[j+1]);
			lengths[(i/2) * Count + (j/2)] = len;
			lengths[(j/2) * Count + (i/2)] = len;
		}
	}*/
}

function set_level(lev) {
	//TODO: transition stuff.
	mode = ModePlay;

	surfWave = Math.random();

	chimeOrder = [0,1,2,3];
	for (var i = 0; i < 4; ++i) {
		var sel = i + (Math.random() * (chimeOrder.length - i)) | 0;
		var temp = chimeOrder[i];
		chimeOrder[i] = chimeOrder[sel];
		chimeOrder[sel] = temp;
	}
	chimeOrder.push(chimeOrder[0]);
	chimeOrder.push(chimeOrder[1]);
	chimeOrder.push(chimeOrder[2]);
	chimeOrder.push(chimeOrder[3]);
	chimeOrder[Math.random() * 8 | 0] = (Math.random() * 4 | 0) + 4;

	//-----------------------------------------------
	level = lev;

	chimes = [];
	chimePairs = [];
	for (var g1 = 0; g1 < level.goals.length; ++g1) {
		for (var g2 = g1 + 1; g2 < level.goals.length; ++g2) {
			chimePairs.push([g1,g2]);
		}
	}
	chimeWait = 4.0 + Math.random();

	level.contours.forEach(function(contour){
		delete contour.attribs;

		var attribs = [];

		var acc = Math.random();

		contour.loops.forEach(function(loop){
			var curSurf = [];
			function push_surf() {
				if (curSurf.length > 8) {
					if (attribs.length > 0) {
						attribs.push(attribs[attribs.length-4], attribs[attribs.length-3], attribs[attribs.length-2], 0.0);
						attribs.push(curSurf[0], curSurf[1], curSurf[2], 0.0);
					}
					attribs.push.apply(attribs, curSurf);
				}
				curSurf = [];
			}
			var prev = { x: loop[loop.length - 2], y: loop[loop.length - 1] };
			for (var i = 0; i + 1 < loop.length; i += 2) {
				var cur = { x: loop[i], y: loop[i+1] };
				var lenSq = (prev.x - cur.x) * (prev.x - cur.x) + (prev.y - cur.y) * (prev.y - cur.y);
				acc += Math.sqrt(lenSq) * 0.01;
				if (lenSq < 0.5 * 0.5) {
					curSurf.push(cur.x, cur.y, acc, 1.0);
				} else {
					push_surf();
					if (curSurf.length > 0) throw "Variables don't work that way";
				}
				prev = cur;
			}
			push_surf();
		});
		//console.log("Have " + attribs.length + " attribs");
		contour.attribs = new Float32Array(attribs);
		//console.log("Contour " + contour.idx + " has " + contour.surfaces.length + " surfaces");
	});

	if (collisionFb == null) {
		collisionTexture = gl.createTexture();
		gl.bindTexture(gl.TEXTURE_2D, collisionTexture);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, canvas.width, canvas.height, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
		gl.bindTexture(gl.TEXTURE_2D, null);
		collisionFb = gl.createFramebuffer();
		gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
		gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, collisionTexture, 0);

		var collisionRb = gl.createRenderbuffer();
		gl.bindRenderbuffer(gl.RENDERBUFFER, collisionRb);
		gl.renderbufferStorage(gl.RENDERBUFFER, gl.DEPTH_STENCIL, canvas.width, canvas.height);
		gl.bindRenderbuffer(gl.RENDERBUFFER, null);
		gl.framebufferRenderbuffer(gl.FRAMEBUFFER, gl.DEPTH_STENCIL_ATTACHMENT, gl.RENDERBUFFER, collisionRb);
	}
	gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
	gl.disable(gl.BLEND);
	gl.clearColor(0.0, 0.0, 0.0, 0.0);
	gl.clear(gl.COLOR_BUFFER_BIT);

	gl.useProgram(solidProgram);
	setCollisionCameraUniforms();


	var vertexBuffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);

	gl.enableVertexAttribArray(solidProgram.aVertex);
	gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 0, 0);

	gl.stencilMask(0xff);
	gl.clear(gl.STENCIL_BUFFER_BIT);
	gl.enable(gl.STENCIL_TEST);
	var idx = 0;
	level.contours.forEach(function(contour){
		contour.idx = idx++;
		var mask = 1 << contour.idx;

		gl.stencilMask(mask);
		gl.stencilFunc(gl.ALWAYS, 0xff, mask);
		gl.colorMask(false, false, false, false);
		gl.stencilOp(gl.KEEP, gl.INVERT, gl.INVERT);
		gl.vertexAttrib4f(solidProgram.aColor, 0.0, 0.0, 0.0, 0.0);
		contour.loops.forEach(function(chain){
			gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( chain ), gl.STREAM_DRAW);
			gl.drawArrays(gl.TRIANGLE_FAN, 0, chain.length / 2);
		});
		gl.colorMask(true, true, true, true);
		gl.stencilFunc(gl.EQUAL, 0xff, mask);
		gl.stencilOp(gl.KEEP, gl.KEEP, gl.KEEP);
		var val = (1 << contour.idx) / 255.0;
		gl.vertexAttrib4f(solidProgram.aColor, val, val, val * 16, 1.0);
		contour.loops.forEach(function(chain){
			gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( chain ), gl.STREAM_DRAW);
			gl.drawArrays(gl.TRIANGLE_FAN, 0, chain.length / 2);
		});

	});
	gl.disable(gl.STENCIL_TEST);

	gl.disableVertexAttribArray(solidProgram.aVertex);
	gl.deleteBuffer(vertexBuffer);

	gl.bindFramebuffer(gl.FRAMEBUFFER, null);

	level.goalsActive = 0;

	level.goals.forEach(function(goal){
		init_goal(goal);
	});

	level.launches = 0;

	check_goals();
}

function check_goals() {
	gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
	buffer = new Uint8Array(4);
	level.goals.forEach(function(goal){
		if (goal.state !== GoalInactive) {
			return;
		}
		var pixelX = goal.origin.x / 35.0 * 350.0 + 350.0;
		var pixelY = goal.origin.y / 15.0 * 150.0 + 150.0;
		pixelX = (pixelX + 0.5) | 0;
		pixelY = (pixelY + 0.5) | 0;
		if (pixelX < 0 || pixelX >= 700 || pixelY < 0 || pixelY >= 300) {
			console.log("Goal with bad center");
		} else {
			gl.readPixels(pixelX, pixelY, 1, 1, gl.RGBA, gl.UNSIGNED_BYTE, buffer);
			if (buffer[0] == 0) {
				playSound(sounds.activates);
				goal.state = GoalSpinning;
				goal.tween = 0.0;
				goal.glow = {rad:0.0, alpha:0.2, cycle:Math.random()};
			}
		}
	});
	gl.bindFramebuffer(gl.FRAMEBUFFER, null);
}

function explode(x,y,rad,idx) {
	var mask = ((2 << idx) - 1);

	gl.useProgram(solidProgram);
	setCollisionCameraUniforms();

	gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
	gl.disable(gl.BLEND);
	gl.enable(gl.STENCIL_TEST);
	gl.stencilMask(mask);

	var verts = new Float32Array( 32 * 2 );
	for (var i = 0; i < 32; ++i) {
		var ang = i / 32.0 * 2.0 * Math.PI;
		verts[2*i] = Math.cos(ang) * rad + x;
		verts[2*i+1] = Math.sin(ang) * rad + y;
	}

	var vertexBuffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);

	gl.enableVertexAttribArray(solidProgram.aVertex);
	gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 0, 0);

	gl.enable(gl.STENCIL_TEST);
	gl.stencilFunc(gl.EQUAL, 0, 0xff ^ mask);
	gl.stencilOp(gl.KEEP, gl.ZERO, gl.ZERO);

	gl.vertexAttrib4f(solidProgram.aColor, 0, 0, 0, 0.0);
	gl.bufferData(gl.ARRAY_BUFFER, verts, gl.STREAM_DRAW);
	gl.drawArrays(gl.TRIANGLE_FAN, 0, verts.length / 2);

	gl.disable(gl.STENCIL_TEST);

	gl.disableVertexAttribArray(solidProgram.aVertex);
	gl.deleteBuffer(vertexBuffer);

	gl.bindFramebuffer(gl.FRAMEBUFFER, null);

	var radSq = rad * rad;
	level.contours.forEach(function(contour){
		if (contour.idx > idx) return;
		for (var i = 0; i + 3 < contour.attribs.length; i += 4) {
			var cx = contour.attribs[i];
			var cy = contour.attribs[i+1];
			var disSq = (cx - x) * (cx - x) + (cy - y) * (cy - y);
			if (disSq < radSq) {
				contour.attribs[i+3] *= disSq / radSq;
			}
		}
	});

	check_goals();
}

//Helper; will switch to solidProgram.
function draw_circle(x,y,rad,  r,g, b,a ) {
	gl.useProgram(solidProgram);
	var verts = new Float32Array( 32 * 2 );
	for (var i = 0; i < 32; ++i) {
		var ang = i / 32.0 * 2.0 * Math.PI;
		verts[2*i] = Math.cos(ang) * rad + x;
		verts[2*i+1] = Math.sin(ang) * rad + y;
	}

	var buffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
	gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( verts ), gl.STREAM_DRAW);

	gl.vertexAttrib4f(solidProgram.aColor, r, g, b, a);
	gl.enableVertexAttribArray(solidProgram.aVertex);
	gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 0, 0);

	gl.drawArrays(gl.TRIANGLE_FAN, 0, verts.length / 2);

	gl.disableVertexAttribArray(solidProgram.aVertex);

	gl.deleteBuffer(buffer);
}


function draw_goals() {
	level.goals.forEach(function(goal){
		if (goal.glow !== undefined) {
			var r1 = Math.sin(goal.glow.cycle * 2.0 * Math.PI) * 0.4 + 1.2;
			var r2 = Math.sin(goal.glow.cycle * 2.0 * Math.PI * 3.0) * 0.3 + 1.3;
			r1 *= goal.glow.rad;
			r2 *= goal.glow.rad;
			draw_circle(goal.at.x,goal.at.y,r1,  1.0,1.0,1.0,goal.glow.alpha);
			draw_circle(goal.at.x,goal.at.y,r2,  1.0,1.0,1.0,goal.glow.alpha);
		}
	});

	var buffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, buffer);

	gl.useProgram(solidProgram);

	gl.enableVertexAttribArray(solidProgram.aVertex);
	gl.vertexAttrib4f(solidProgram.aColor, 1.0, 1.0, 1.0, 0.9);

	level.goals.forEach(function(goal){
		if (goal.lines !== undefined && (goal.state === GoalInactive || goal.state === GoalSpinning)) {
			var mat = multMatrix(getCameraMatrix(), getScaleTranslateMatrix(GoalRad, goal.at.x, goal.at.y, 0.0));
			mat = multMatrix(mat, getRotateMatrix(goal.x_dir, goal.y_dir, goal.z_dir));
			gl.uniformMatrix4fv(gl.getUniformLocation(gl.getParameter(gl.CURRENT_PROGRAM), "Matrix"), false, mat);
			gl.bufferData(gl.ARRAY_BUFFER, goal.lines, gl.STREAM_DRAW);
			gl.vertexAttribPointer(solidProgram.aVertex, 3, gl.FLOAT, false, 3 * 4, 0);
			gl.drawArrays(gl.LINES, 0, goal.lines.length / 3);
		}
	});

	gl.disableVertexAttribArray(solidProgram.aVertex);

	//--------------------------

	gl.useProgram(particleProgram);
	setCameraUniforms();
	gl.enableVertexAttribArray(particleProgram.aXabcd);
	gl.enableVertexAttribArray(particleProgram.aYabcd);
	gl.uniform4f(particleProgram.uColor, 1.0, 1.0, 1.0, 1.0);
	gl.vertexAttribPointer(particleProgram.aXabcd, 4, gl.FLOAT, false, 4 * 8, 0);
	gl.vertexAttribPointer(particleProgram.aYabcd, 4, gl.FLOAT, false, 4 * 8, 4 * 4);
	gl.vertexAttrib2f(particleProgram.aTScaleOfs, 1.0, 0.0);

	level.goals.forEach(function(goal){
		if (goal.explodeAttribs === undefined) return;
		gl.uniform1f(particleProgram.uT, goal.tween);
		gl.bufferData(gl.ARRAY_BUFFER, goal.explodeAttribs, gl.STREAM_DRAW);
		gl.drawArrays(gl.LINES, 0, goal.explodeAttribs.length / 8);
	});
	gl.disableVertexAttribArray(particleProgram.aXabcd);
	gl.disableVertexAttribArray(particleProgram.aYabcd);

	//--------------------------

	gl.useProgram(solidProgram);
	setCameraUniforms();
	gl.enableVertexAttribArray(solidProgram.aVertex);

	var verts = [];
	var s = 0.3;
	level.goals.forEach(function(goal){
		if (goal.state === GoalActive) {
			verts.push(goal.at.x - s, goal.at.y);
			verts.push(goal.at.x + s, goal.at.y);
			verts.push(goal.at.x, goal.at.y - s);
			verts.push(goal.at.x, goal.at.y + s);
		} else {
		/*
			verts.push(goal.at.x - s, goal.at.y - s);
			verts.push(goal.at.x + s, goal.at.y - s);
			verts.push(goal.at.x + s, goal.at.y - s);
			verts.push(goal.at.x + s, goal.at.y + s);
			verts.push(goal.at.x + s, goal.at.y + s);
			verts.push(goal.at.x - s, goal.at.y + s);
			verts.push(goal.at.x - s, goal.at.y + s);
			verts.push(goal.at.x - s, goal.at.y - s);
		*/
		}
	});

	gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( verts ), gl.STREAM_DRAW);
	gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 0, 0);

	gl.vertexAttrib4f(solidProgram.aColor, 1.0, 1.0, 1.0, 0.9);
	gl.drawArrays(gl.LINES, 0, verts.length / 2);

	gl.disableVertexAttribArray(solidProgram.aVertex);

	gl.deleteBuffer(buffer);
}

function draw_followers() {
	followers.forEach(function(f){
		draw_circle(f.x, f.y, f.rad,   1.0, 0.9, 0.96, 0.6 * f.alpha);
	});
}

function draw_level() {
	gl.useProgram(solidProgram);
	setCameraUniforms();

	var buffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, buffer);

	//---------------------------------------------
	//draw collision texture:

	gl.useProgram(levelbgProgram);
	setCameraUniforms();
	gl.uniform1i(gl.getUniformLocation(levelbgProgram, "tex"), 0);
	gl.uniform1i(gl.getUniformLocation(levelbgProgram, "noise"), 1);

	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "NoiseMul"), canvas.width / 128, canvas.height / 128);

	var attribs = [
		-35,-15, 0,0,
		-35, 15, 0,1,
		 35,-15, 1,0,
		 35, 15, 1,1,
	];

	gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( attribs ), gl.STREAM_DRAW);

	gl.enableVertexAttribArray(levelbgProgram.aVertex);
	gl.enableVertexAttribArray(levelbgProgram.aTexCoord);
	gl.vertexAttribPointer(levelbgProgram.aVertex, 2, gl.FLOAT, false, 4 * 4, 0);
	gl.vertexAttribPointer(levelbgProgram.aTexCoord, 2, gl.FLOAT, false, 4 * 4, 4 * 2);

	gl.activeTexture(gl.TEXTURE1);
	gl.bindTexture(gl.TEXTURE_2D, noiseTexture);
	gl.activeTexture(gl.TEXTURE0);
	gl.bindTexture(gl.TEXTURE_2D, collisionTexture);

	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "noiseScale"), 15.0 / canvas.width, 15.0 / canvas.height);
	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "NoiseOfs"), Math.random(), Math.random());
	gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "noiseScale"), 5.0 / canvas.width, 3.0 / canvas.height);
	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "NoiseOfs"), Math.random(), Math.random());
	gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "noiseScale"), 3.0 / canvas.width, 3.0 / canvas.height);
	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "NoiseOfs"), Math.random(), Math.random());
	gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "NoiseOfs"), Math.random(), Math.random());
	gl.uniform2f(gl.getUniformLocation(levelbgProgram, "noiseScale"), 2.0 / canvas.width, 2.0 / canvas.height);
	gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

	gl.activeTexture(gl.TEXTURE1);
	gl.bindTexture(gl.TEXTURE_2D, null);
	gl.activeTexture(gl.TEXTURE0);
	gl.bindTexture(gl.TEXTURE_2D, null);

	gl.disableVertexAttribArray(levelbgProgram.aVertex);
	gl.disableVertexAttribArray(levelbgProgram.aTexCoord);


	
/* DEBUG: show collisionFb
	gl.useProgram(textureProgram);
	setCameraUniforms();
	gl.uniform1i(gl.getUniformLocation(textureProgram, "tex"), 0);

	var attribs = [
		-35,-15, 0,0,
		-35, 15, 0,1,
		 35,-15, 1,0,
		 35, 15, 1,1,
	];

	var buffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
	gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( attribs ), gl.STREAM_DRAW);

	gl.enableVertexAttribArray(textureProgram.aVertex);
	gl.enableVertexAttribArray(textureProgram.aTexCoord);
	gl.vertexAttribPointer(textureProgram.aVertex, 2, gl.FLOAT, false, 4 * 4, 0);
	gl.vertexAttribPointer(textureProgram.aTexCoord, 2, gl.FLOAT, false, 4 * 4, 4 * 2);

	gl.bindTexture(gl.TEXTURE_2D, collisionTexture);

	gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

	gl.bindTexture(gl.TEXTURE_2D, null);

	gl.disableVertexAttribArray(textureProgram.aVertex);
	gl.disableVertexAttribArray(textureProgram.aTexCoord);

	gl.deleteBuffer(buffer);
	*/

	//---------------------------------------------
	//draw surfaces:

	gl.useProgram(surfaceProgram);
	setCameraUniforms();
	gl.uniform1i(surfaceProgram.uTex, 0);

	gl.enableVertexAttribArray(surfaceProgram.aInfo);
	gl.vertexAttribPointer(surfaceProgram.aInfo, 4, gl.FLOAT, false, 4 * 4, 0);

	gl.bindTexture(gl.TEXTURE_2D, surfaceTexture);

	gl.enable(gl.BLEND);
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE);
	var colors = [
		{r:0x41/0xff, g:0x08/0xff, b:0x2f/0xff, a:0.5},
		{r:0x58/0xff, g:0x1a/0xff, b:0x36/0xff, a:0.3},
		{r:0x72/0xff, g:0x6a/0xff, b:0x41/0xff, a:0.3},
		{r:0x27/0xff, g:0x44/0xff, b:0x3c/0xff, a:0.3},
		{r:0x1e/0xff, g:0x4a/0xff, b:0x56/0xff, a:0.3},
	];
	level.contours.forEach(function(contour){
		gl.bufferData(gl.ARRAY_BUFFER, contour.attribs, gl.STREAM_DRAW);

		var color = colors[contour.idx % colors.length];

		gl.uniform2f(surfaceProgram.uNoiseScale, 1.0, 1.0);
		gl.uniform4f(surfaceProgram.uColor, color.r, color.g * 0.9, color.b * 0.9, color.a);
		gl.uniform2f(surfaceProgram.uScaleOfs, 1.0, 3.0 * surfWave + 0.0);
		gl.drawArrays(gl.LINE_STRIP, 0, contour.attribs.length / 4);

		gl.uniform4f(surfaceProgram.uColor, color.r * 0.9, color.g, color.b * 0.9, color.a);
		gl.uniform2f(surfaceProgram.uScaleOfs, 0.4, 4.0 * surfWave + 0.2);
		gl.drawArrays(gl.LINE_STRIP, 0, contour.attribs.length / 4);

		gl.uniform4f(surfaceProgram.uColor, color.r * 0.9, color.g * 0.9, color.b, color.a);
		gl.uniform2f(surfaceProgram.uScaleOfs, 1.5, -5.0 * surfWave);
		gl.drawArrays(gl.LINE_STRIP, 0, contour.attribs.length / 4);
	});
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

	gl.bindTexture(gl.TEXTURE_2D, null);

	gl.disableVertexAttribArray(surfaceProgram.aInfo);

	gl.deleteBuffer(buffer);

	//-----------------------------------------------------
	//Chimes:
	(function(){
		var attribs = [];
		chimes.forEach(function(chime){
			attribs.push(chime.a.x, chime.a.y);
			attribs.push(1.0, 1.0, 1.0, chime.amt);
			attribs.push(chime.b.x, chime.b.y);
			attribs.push(1.0, 1.0, 1.0, chime.amt);
		});

		gl.useProgram(solidProgram);
		var buffer = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
		gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( attribs ), gl.STREAM_DRAW);

		gl.enableVertexAttribArray(solidProgram.aVertex);
		gl.enableVertexAttribArray(solidProgram.aColor);
		gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 4 * 6, 0);
		gl.vertexAttribPointer(solidProgram.aColor, 4, gl.FLOAT, false, 4 * 6, 4 * 2);

		gl.drawArrays(gl.LINES, 0, attribs.length / 6);

		gl.disableVertexAttribArray(solidProgram.aVertex);
		gl.disableVertexAttribArray(solidProgram.aColor);

		gl.deleteBuffer(buffer);

	})();


	//-----------------------------------------------------
	//Show goals:
	draw_goals();

	//-----------------------------------------------------
	//Cursor / followers:
	draw_followers();

	//-----------------------------------------------------
	//Show projectile:
	if (projectile !== undefined) {
		gl.useProgram(solidProgram);
		//setCameraUniforms(); //already called above.

		var verts = new Float32Array( 32 * 2 );
		for (var i = 0; i < 32; ++i) {
			var ang = i / 32.0 * 2.0 * Math.PI;
			verts[2*i] = Math.cos(ang) * ProjectileRadius + projectile.x;
			verts[2*i+1] = Math.sin(ang) * ProjectileRadius + projectile.y;
		}

		var buffer = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
		gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( verts ), gl.STREAM_DRAW);

		gl.vertexAttrib4f(solidProgram.aColor, 1.0, 1.0, 1.0, 0.9);
		gl.enableVertexAttribArray(solidProgram.aVertex);
		gl.vertexAttribPointer(solidProgram.aVertex, 2, gl.FLOAT, false, 0, 0);

		gl.drawArrays(gl.TRIANGLE_FAN, 0, verts.length / 2);

		gl.disableVertexAttribArray(solidProgram.aVertex);

		gl.deleteBuffer(buffer);
	}
	
}

function draw_title() {
	if (titleTexture === null) return;
	gl.useProgram(textureProgram);
	setCameraUniforms();
	gl.uniform1i(gl.getUniformLocation(textureProgram, "tex"), 0);

	var buffer = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
	
	gl.enableVertexAttribArray(textureProgram.aVertex);
	gl.enableVertexAttribArray(textureProgram.aTexCoord);
	gl.vertexAttribPointer(textureProgram.aVertex, 2, gl.FLOAT, false, 4 * 4, 0);
	gl.vertexAttribPointer(textureProgram.aTexCoord, 2, gl.FLOAT, false, 4 * 4, 4 * 2);

	(function(){
		var l = ((700 - titleTexture.width) / 2) | 0;
		var t = 270;
		l = (l - 350) / 350 * 35;
		t = (t - 150) / 150 * 15;
		var w = titleTexture.width / 700 * 70;
		var h = titleTexture.height / 300 * 30;

		var attribs = [
			l,   t, 0,0,
			l,   t-h,   0,1,
			l+w, t, 1,0,
			l+w, t-h,   1,1
		];

		gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( attribs ), gl.STREAM_DRAW);

		gl.bindTexture(gl.TEXTURE_2D, titleTexture);

		gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

		gl.bindTexture(gl.TEXTURE_2D, null);
	})();

	function well_done(){
		if (wellDoneTexture === null) return;
		var l = ((700 - wellDoneTexture.width) / 2) | 0;
		var t = 55;
		l = (l - 350) / 350 * 35;
		t = (t - 150) / 150 * 15;
		var w = wellDoneTexture.width / 700 * 70;
		var h = wellDoneTexture.height / 300 * 30;

		var attribs = [
			l,   t, 0,0,
			l,   t-h,   0,1,
			l+w, t, 1,0,
			l+w, t-h,   1,1
		];

		gl.bufferData(gl.ARRAY_BUFFER, new Float32Array( attribs ), gl.STREAM_DRAW);

		gl.bindTexture(gl.TEXTURE_2D, wellDoneTexture);

		gl.drawArrays(gl.TRIANGLE_STRIP, 0, attribs.length / 4);

		gl.bindTexture(gl.TEXTURE_2D, null);
	}

	var win = true;
	for (var i = 0; i < levels.length; ++i) {
		if (scores[i] === undefined) {
			win = false;
		}
	}

	if (win) well_done();

	gl.disableVertexAttribArray(textureProgram.aVertex);
	gl.disableVertexAttribArray(textureProgram.aTexCoord);

	gl.deleteBuffer(buffer);
}

function draw() {
	gl.clearColor(0.0, 0.0, 0.0, 1.0);
	gl.clear(gl.COLOR_BUFFER_BIT);

	gl.enable(gl.BLEND);
	gl.blendEquation(gl.FUNC_ADD);
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

	if (mode === ModePlay || mode === ModeToSelect) {
		draw_level();
	} else if (mode === ModeSelect) {
		draw_title();
		draw_goals();
		draw_followers();
	}
}


function updateProjectile() {
	projectile.x += Tick * projectile.vx;
	projectile.y += Tick * projectile.vy;
	projectile.vy -= Tick * Gravity;
	var pixelX = projectile.x / 35.0 * 350.0 + 350.0;
	var pixelY = projectile.y / 15.0 * 150.0 + 150.0;
	var pixelRadius = ProjectileRadius / 35.0 * 350.0;
	var pixelMinX = pixelX - pixelRadius;
	var pixelMaxX = pixelX + pixelRadius;
	var pixelMinY = pixelY - pixelRadius;
	var pixelMaxY = pixelY + pixelRadius;

	if (pixelMaxX < 0.0 || pixelMinX > 700.0 || pixelMaxY < 0.0) {
		projectile = undefined;
		return;
	}

	var readMinX = Math.max(pixelMinX | 0, 0);
	var readMaxX = Math.min(pixelMaxX | 0 + 1, 700 - 1);
	var readMinY = Math.max(pixelMinY | 0, 0);
	var readMaxY = Math.min(pixelMaxY | 0 + 1, 300 - 1);
	var readWidth = readMaxX - readMinX + 1;
	var readHeight = readMaxY - readMinY + 1;
	if (readWidth <= 0 || readHeight <= 0) return;
	//console.log("Readback is [" + readMinX + ", " + readMaxX + "]x[" + readMinY + ", " + readMaxY + "]");
	gl.bindFramebuffer(gl.FRAMEBUFFER, collisionFb);
	var pixels = new Uint8Array(readWidth * readHeight * 4);
	gl.readPixels(readMinX, readMinY, readWidth, readHeight, gl.RGBA, gl.UNSIGNED_BYTE, pixels);
	gl.bindFramebuffer(gl.FRAMEBUFFER, null);
	var didHit = undefined;
	for (var py = 0; py < readHeight; ++py) {
		for (var px = 0; px < readWidth; ++px) {
			var val = pixels[(py * readWidth + px) * 4 + 0];
			if (val != 0) {
				var c = {
					x: ((px + 0.5 + readMinX) - 350.0) / 350.0 * 35.0,
					y: ((py + 0.5 + readMinY) - 150.0) / 150.0 * 15.0
				};
				var disSq = (c.x - projectile.x) * (c.x - projectile.x) + (c.y - projectile.y) * (c.y - projectile.y);
				if (disSq < ProjectileRadius * ProjectileRadius) {
					didHit = val;
					break;
				}
			}
		}
		if (didHit !== undefined) break;
	}
	if (didHit !== undefined) {
		var idx = 0;
		while ((didHit & (1 << idx)) == 0) {
			++idx;
		}
		//console.log("Hit: " + idx.toString(16));
		playSound(sounds.explodes);
		if (sounds.release !== null) {
			sounds.release.volume = 0.0;
		}
		explode(projectile.x, projectile.y, 5.0, idx);

		projectile = undefined;
	}
}

var followScale = 0.1;

function update_followers(elapsed) {

	function follow_t1() {
		if (mode === ModePlay && projectile !== undefined && projectile.launched < 1.0) {
			return {x:projectile.x, y:projectile.y};
		}
		return {x:mouse.x, y:mouse.y};
	}
	function follow_t0() {
		if (mode === ModePlay) {
			if (projectile !== undefined) {
				if (projectile.launched < 1.0) {
					return {x:nearGoal.at.x, y:nearGoal.at.y};
				} else {
					return {x:mouse.x, y:mouse.y};
				}
			} else if (nearGoal !== undefined) {
				return {x:nearGoal.at.x, y:nearGoal.at.y};
			}
		} else if (mode === ModeSelect) {
			if (nearGoal !== undefined) {
				return {x:nearGoal.at.x, y:nearGoal.at.y};
			}
		}
		return {x:mouse.x, y:mouse.y};
	}

	var targetScale = followScale;
	if (projectile === undefined) {
		targetScale = 1.0;
	} else if (projectile !== undefined && projectile.launched < 1.0) {
		targetScale = 0.7;
	} else {
		targetScale = 0.1;
	}
	if (mode === ModeToSelect) {
		targetScale = 0.0;
	}
	followScale += (targetScale - followScale) * 0.1;

	followers = followers.filter(function(f){
		if (f == followers[0]) {
			f.x = mouse.x;
			f.y = mouse.y;
			f.t = 1.0;
			f.rad = 0.2;
			f.alpha = 1.0;
			return true;
		}

		var t0 = follow_t0();
		var t1 = follow_t1();
		var t = {x:t0.x + f.t * (t1.x - t0.x), y:t0.y + f.t * (t1.y - t0.y)};

		if (isNaN(t.x)) throw "PROBLEM";
		if (isNaN(t.y)) throw "PROBLEM";

		f.along += Tick * f.invLife;
		if (f.along < 0.3) {
			f.rad = f.maxRad * f.along / 0.3;
		} else if (f.along < 0.8) {
			f.rad = f.maxRad;
		} else {
			f.rad = f.maxRad * (1.0 - f.along) / 0.2;
		}
		f.rad *= followScale;


		var damp = 0.99;
		f.vx *= damp;
		f.vy *= damp;

		f.vx += (t.x - f.x) * Tick * 50.0;
		f.vy += (t.y - f.y) * Tick * 50.0;

		f.x += f.vx * Tick;
		f.y += f.vy * Tick;

		f.alpha = 1.0;
		var len = Math.sqrt((t.x - f.x) * (t.x - f.x) + (t.y - f.y) * (t.y - f.y));
		if (len > 1.0) {
			f.alpha -= (len - 1.0);
			if (f.alpha < 0) {
				f.alpha = 0;
				return false;
			}
		}

		return f.along < 1.0;
	});

	if (followers.length < 10) {
		var t = Math.random();
		var t1 = follow_t1();
		var t0 = follow_t0();
		var tv = {x: t0.x - t1.x, y: t0.y - t1.y};
		var tvs = (Math.random() - 0.5) * 3.0;
		tv.x *= tvs;
		tv.y *= tvs;
		followers.push({
			x: t0.x + (t1.x - t0.x) * t + (Math.random() - 0.5) * 0.4,
			y: t0.y + (t1.y - t0.y) * t + (Math.random() - 0.5) * 0.4,
			vx: (Math.random() - 0.5) * 5.0 + tv.x,
			vy: (Math.random() - 0.5) * 5.0 + tv.y,
			t: t,
			rad: 0.0,
			maxRad: 0.3 + Math.random() * 0.2,
			invLife: 1.0 / (0.2 + 0.6 * Math.random()),
			along: 0.0,
		});
	}
}

function record_score() {
	if (level.goalsActive != level.goals.length) {
		console.log("Trying to score without having finished.");
		return;
	}
	if (level.launches === undefined) {
		console.log("Trying to score without knowing how many launches.");
		level.launches = 100;
	}
	for (var i = 0; i < levels.length; ++i) {
		if (levels[i] === level) {
			if (scores[i] === undefined) {
				scores[i] = { launches: level.launches };
			} else if (scores[i].launches > level.launches) {
				scores[i].launches = level.launches;
			}
			//console.log("Have " + scores[i].launches + " for " + i);
		}
	}
	write_scores();
}

function write_scores() {
	for (var i = 0; i < levels.length; ++i) {
		if (scores[i] === undefined) {
			localStorage.removeItem("score_" + levels[i].name);
		} else {
			localStorage.setItem("score_" + levels[i].name, scores[i].launches);
		}
	}
}

function read_scores() {
	scores = [];
	for (var i = 0; i < levels.length; ++i) {
		scores[i] = undefined;
		var item = localStorage.getItem("score_" + levels[i].name);
		if (item !== null) {
			var launches = parseInt(item);
			if (isFinite(launches)) {
				scores[i] = {launches:launches};
			}
		}
	}
}

function delete_scores() {
	scores = [];
	write_scores();
	read_scores();
	set_select();
}

var muted = false;

function mute() {
	if (muted) {
		muted = false;
		localStorage.removeItem("mute");
		document.getElementById("mute").innerText = "mute"
		if (sounds.background) {
			sounds.background.volume = BgVolume;
		}
	} else {
		muted = true;
		localStorage.setItem("mute", "mute");
		document.getElementById("mute").innerText = "unmute"
		for (var n in sounds) {
			var s = sounds[n];
			if (s instanceof Array) {
				s.forEach(function(p){
					if (p !== null) p.volume = 0.0;
				});
			} else if (s !== null) {
				s.volume = 0.0;
			}
		}
	}
}

var goalRot = Math.random();

function update_play(elapsed) {
	if (projectile !== undefined) {
		if (projectile.launched == 0.0) {
			//nothing much to do.
		} else if (projectile.launched < 1.0) {
			var launchTime = (Math.PI / 2) / projectile.launchF;
			projectile.launched += Tick / launchTime;
			if (projectile.launched >= 1.0) {
				projectile.launched = 1.0;
				projectile.x = nearGoal.at.x;
				projectile.y = nearGoal.at.y;
			} else {
				var amt = Math.sin(Math.PI / 2 * (projectile.launched - 1.0));
				projectile.x = nearGoal.at.x + projectile.launchTo.x * amt;
				projectile.y = nearGoal.at.y + projectile.launchTo.y * amt;
				//if (isNaN(projectile.x)) throw new Exception();
			}
		} else {
			updateProjectile();
		}
	}

	goalRot += Tick;
	if (goalRot > 2.0 * Math.PI) goalRot -= 2.0 * Math.PI;

	level.goals.forEach(function(goal){
		if (goal.state === GoalSpinning || goal.state === GoalExploding) {
			goal.glow.cycle += elapsed / 0.5;
			if (goal.glow.cycle > 1.0) goal.glow.cycle -= 1.0;
		} else if (goal.state === GoalActive) {
			goal.glow.cycle += elapsed / 1.0;
			if (goal.glow.cycle > 1.0) goal.glow.cycle -= 1.0;
		}
		if (goal.state === GoalSpinning) {
			goal.tween += elapsed / 1.2;
			if (goal.tween < 1.0) {
				var amt = goal.tween;
				amt *= amt;
				var s = amt * 50.0 + 1.0;
				var fy = Math.sin(goalRot) * s * Tick;
				var fz = Math.cos(goalRot) * s * Tick;
				goal.x_dir.x += fy * goal.y_dir.x;
				goal.x_dir.y += fy * goal.y_dir.y;
				goal.x_dir.z += fy * goal.y_dir.z;
				goal.x_dir.x += fz * goal.z_dir.x;
				goal.x_dir.y += fz * goal.z_dir.y;
				goal.x_dir.z += fz * goal.z_dir.z;
				goal.x_dir = normalize3(goal.x_dir);
				var d = dot3(goal.x_dir, goal.y_dir);
				goal.y_dir.x -= goal.x_dir.x * d;
				goal.y_dir.y -= goal.x_dir.y * d;
				goal.y_dir.z -= goal.x_dir.z * d;
				goal.y_dir = normalize3(goal.y_dir);
				goal.z_dir = cross3(goal.x_dir, goal.y_dir);

				goal.glow.rad = goal.tween;
			} else {
				goal.state = GoalExploding;
				goal.tween = 0.0;
				goal.glow.rad = 1.0;
				var attribs = new Float32Array((goal.lines.length / 3) * 8);
				var outVel = {x:0, y:0};
				var inVel = {x:0, y:0};
				var dir = {x:0, y:0};
				for (var i = 0; i + 2 < goal.lines.length; i += 3) {
					dir = {x:-dir.x, y:-dir.y};
					if ((i % 6) == 0) {
						outVel = normalize2({x: Math.random() - 0.5, y: Math.random() - 0.5});
						inVel = normalize2({x: Math.random() - 0.5, y: Math.random() - 0.5});
						outVel.x *= 5;
						outVel.y *= 5;
						inVel.x *= 5;
						inVel.y *= 5;
						var ang = i / goal.lines.length * Math.PI;
						dir = {x: Math.cos(ang), y: Math.sin(ang) };
						var len = 0.3 + 0.1 * Math.random();
						dir.x *= len;
						dir.y *= len;
					}
					var pt = {x:goal.lines[i], y:goal.lines[i+1], z:goal.lines[i+2]};
					pt = {
						x: goal.x_dir.x * pt.x + goal.y_dir.x * pt.y + goal.z_dir.x * pt.z,
						y: goal.x_dir.y * pt.x + goal.y_dir.y * pt.y + goal.z_dir.y * pt.z
					};
					pt.x = GoalRad * pt.x + goal.origin.x;
					pt.y = GoalRad * pt.y + goal.origin.y;
					var a = (i / 3) * 8;
					attribs[a + 0] = pt.x;
					attribs[a + 1] = pt.x + outVel.x;
					attribs[a + 2] = goal.target.x + dir.x - inVel.x;
					attribs[a + 3] = goal.target.x + dir.x;
					attribs[a + 4] = pt.y;
					attribs[a + 5] = pt.y + outVel.y;
					attribs[a + 6] = goal.target.y + dir.y - inVel.y;
					attribs[a + 7] = goal.target.y + dir.y;
				}
				goal.explodeAttribs = attribs;
			}
		} else if (goal.state === GoalExploding) {
			goal.tween += elapsed / 0.8;
			if (goal.tween < 1.0) {
				var amt = goal.tween;
				/*
				goal.at = {x: goal.origin.x + amt * (goal.target.x - goal.origin.x),
						y: goal.origin.y + amt * (goal.target.y - goal.origin.y)};
						*/
				goal.glow.alpha = 0.2 * (1.0 - amt);
				goal.glow.rad = 1.0 + 2.0 * amt * amt;
			} else {
				goal.state = GoalActive;
				delete goal.tween;
				delete goal.explodeAttribs;
				goal.glow.alpha = 0.1;
				goal.glow.rad = 1.0;
				goal.at = {x: goal.target.x, y: goal.target.y};
			}
			
		}
	});

	level.goalsActive = 0;
	level.goals.forEach(function(goal){
		if (goal.state === GoalActive) {
			level.goalsActive += 1;
		}
	});

	if (level.goalsActive == level.goals.length) {
		if (mode !== ModeToSelect) {
			mode = ModeToSelect;
			projectile = undefined;
			chimePattern = [];
			var beat = 60.0 / BPM;
			var fac = 0.95;
			chimePattern.push({len: 4 * beat * fac, wait: 2 * beat});
			for (var i = 0; i < 2; ++i) {
				chimePattern.push({len: 4 * beat * fac, wait: 2 * beat});
			}
			for (var i = 0; i < 4; ++i) {
				chimePattern.push({len: 2 * beat * fac, wait: 2 * beat});
			}
			for (var i = 0; i < 4; ++i) {
				chimePattern.push({len: beat * fac, wait: beat * beat});
			}
			for (var i = 0; i < 8; ++i) {
				chimePattern.push({len: 4 * beat * fac, wait: 0.0, mute:(i >= 1)});
			}
			chimeWait = 0.0;

			//record the win:
			record_score();
		}
	}

	update_followers();

	chimes = chimes.filter(function(chime){
		chime.along += Tick * chime.invLife;
		if (chime.along >= 1.0) {
			chimePairs.push(chime.pair);
			return false;
		}

		if (chime.along < 0.2) {
			chime.amt = chime.along / 0.2;
		} else {
			chime.amt = (1.0 - chime.along) / (1.0 - 0.2);
		}
		if (chime.loud) {
			chime.amt = 1.0 - (1.0 - chime.amt) * (1.0 - chime.amt);
		} else {
			chime.amt *= chime.amt;
			chime.amt *= 0.2;
		}
		return true;
	});

	chimeWait -= Tick;
	if (chimeWait <= 0.0) {
		chimeWait = 8.0 * 60.0 / BPM;
		var chimeLength = 4.0 * 60.0 / BPM;
		var chimeProbability = 0.5;
		var chimeMute = false;

		if (mode === ModeToSelect) {
			if (chimePattern.length == 0) {
				if (chimes.length == 0) {
					set_select();
					return;
				}
				chimeWait = 0.0;
				chimeProbability =-1.0;
				chimeLength = 0.0;
			} else {
				chimeWait = chimePattern[0].wait;
				chimeLength = chimePattern[0].len;
				if (chimePattern[0].mute !== undefined) {
					chimeMute = chimePattern[0].mute;
				}
				chimeProbability = 1.0;
				chimePattern.shift();
			}
		}

		if ((chimes.length < 5 || mode === ModeToSelect) && chimePairs.length && Math.random() <= chimeProbability) {
			var sel = (Math.random() * chimePairs.length) | 0;
			if (sel >= chimePairs.length) sel = chimePairs.length - 1;
			var pair = chimePairs[sel];
			chimePairs[sel] = chimePairs[chimePairs.length - 1];
			chimePairs.pop();
			var loud = (level.goals[pair[0]].state === GoalActive) && (level.goals[pair[1]].state === GoalActive);
			chimes.push({
				pair:pair,
				a:{x: level.goals[pair[0]].at.x, y: level.goals[pair[0]].at.y},
				b:{x: level.goals[pair[1]].at.x, y: level.goals[pair[1]].at.y},
				amt:0.0,
				loud:loud,
				along:0.0,
				invLife: 1.0/chimeLength
			});
			if (!chimeMute) {
				var volume = 0.5;
				if (loud) volume = 0.9;
				if (chimeLength >= 3.0 * 60.0 / BPM) {
					playSound(sounds.wholeChimes, volume);
				} else if (chimeLength >= 1.5 * 60.0 / BPM) {
					playSound(sounds.halfChimes, volume);
				} else if (chimeLength >= 0.5 * 60.0 / BPM) {
					playSound(sounds.quarterChimes, volume);
				}
			}
		}
	}
}

function update_select(elapsed) {
	update_followers();

	level.goals.forEach(function(goal) {
		if (goal.glow !== undefined) {
			goal.glow.cycle += Tick / (3.0 * (60.0 / BPM));
			if (goal.glow.cycle >= 1.0) {
				goal.glow.cycle -= 1.0;
			}
		}
	});
}

function update(elapsed) {
	while (mouse.prevx.length >= 30) {
		mouse.prevx.shift();
		mouse.prevy.shift();
	}
	mouse.prevx.push(mouse.x);
	mouse.prevy.push(mouse.y);
	update_nearGoal();
	if (mode === ModeSelect) {
		update_select(elapsed);
	} else if (mode === ModePlay || mode == ModeToSelect) {
		update_play(elapsed);
	}
	surfWave += Tick / 100.0;
	if (surfWave > 1.0) surfWave -= 1.0;
}

function loop(timestamp) {
	if (sounds.background !== null && sounds.background.currentTime > 16.0) {
		var step = 10.0;
		if (Math.random() < 0.5) {
			step += 2.0;
		}
		if (Math.random() < 0.5) {
			step += 2.0;
		}
		sounds.background.currentTime -= step;
	}
	var elapsed = 0;
	if (timestamp === undefined) {
		elapsed = 1.0 / 60.0;
	} else {
		if (loop.lastTimestamp) {
			elapsed = (timestamp - loop.lastTimestamp) / 1000.0;
		}
		loop.lastTimestamp = timestamp;
	}
	// TODO: actually pay attention to elapsed

	update(elapsed);

	draw();

	window.requestAnimFrame(loop);
}

function sound(name, src, idx) {
	if (!sounds.hasOwnProperty(name)) throw "No sound name '" + name + "'";

	var a = new Audio();
	var loaded = [0];
	a.addEventListener('canplaythrough', function(){
		if (loaded[0] == 1) return;
		loaded[0] = 1;
		if (sounds[name] instanceof Array) {
			if (idx !== undefined) {
				sounds[name][idx] = a;
			} else {
				sounds[name].push(a);
			}
		} else {
			if (sounds[name] != null) {
				console.log("Sound collision on '" + name + "'");
			}
			sounds[name] = a;
		}
		if (name == "background") {
			sounds.background.loop = true;
			sounds.background.currentTime = 0.0;
			sounds.background.volume = BgVolume;
			sounds.background.play();
		}
		if (muted) {
			a.volume = 0.0;
		}
	});
	a.addEventListener('error', function(){
		console.log("Audio loading error for " + a.src);
	});
	a.src = src;
}

function setup() {
	canvas = document.getElementById("game");

	if (localStorage.getItem("mute")) {
		mute();
	}

	//get into all those audio tags and fix things up:
	
	camera.right = {
		x: camera.up.y * (canvas.width / canvas.height),
		y: -camera.up.x * (canvas.width / canvas.height)
	};

	//---------------------------------------
	//gl init from mozilla's webGL docs:

	gl = null;
	try {
		var attribs = {};
		attribs.antialias = true;
		attribs.stencil = true;
		gl = canvas.getContext("webgl", attribs) || canvas.getContext("experimental-webgl", attribs);
	} catch (e) {}

	if (!gl) {
		alert("Unable to init webGL. Your browser may not support it.");
		return;
	}

	//----------------------------------------
	//Program setup:

	solidProgram = getProgram("solid-vs", "solid-fs");
	if (!solidProgram) throw "Failed to compile solidProgram";
	solidProgram.aVertex = gl.getAttribLocation(solidProgram, "Vertex");
	solidProgram.aColor = gl.getAttribLocation(solidProgram, "Color");

	textureProgram = getProgram("texture-vs", "texture-fs");
	if (!textureProgram) throw "Failed to compile textureProgram";
	textureProgram.aVertex = gl.getAttribLocation(textureProgram, "Vertex");
	textureProgram.aTexCoord = gl.getAttribLocation(textureProgram, "TexCoord");

	levelbgProgram = getProgram("levelbg-vs", "levelbg-fs");
	if (!levelbgProgram) throw "Failed to compile levelbgProgram";
	levelbgProgram.aVertex = gl.getAttribLocation(levelbgProgram, "Vertex");
	levelbgProgram.aTexCoord = gl.getAttribLocation(levelbgProgram, "TexCoord");

	particleProgram = getProgram("particle-vs", "particle-fs");
	if (!particleProgram) throw "Failed to compile particleProgram";
	particleProgram.aXabcd = gl.getAttribLocation(particleProgram, "Xabcd");
	particleProgram.aYabcd = gl.getAttribLocation(particleProgram, "Yabcd");
	particleProgram.aTScaleOfs = gl.getAttribLocation(particleProgram, "TScaleOfs");
	particleProgram.uMatrix = gl.getUniformLocation(particleProgram, "Matrix");
	particleProgram.uT = gl.getUniformLocation(particleProgram, "T");
	particleProgram.uColor = gl.getUniformLocation(particleProgram, "Color");

	surfaceProgram = getProgram("surface-vs", "surface-fs");
	if (!surfaceProgram) throw "Failed to compile surfaceProgram";
	surfaceProgram.aInfo = gl.getAttribLocation(surfaceProgram, "Info");
	surfaceProgram.uMatrix = gl.getUniformLocation(surfaceProgram, "Matrix");
	surfaceProgram.uTex = gl.getUniformLocation(surfaceProgram, "Tex");
	surfaceProgram.uScaleOfs = gl.getUniformLocation(surfaceProgram, "ScaleOfs");
	surfaceProgram.uNoiseScale = gl.getUniformLocation(surfaceProgram, "NoiseScale");
	surfaceProgram.uColor = gl.getUniformLocation(surfaceProgram, "Color");



	//----------------------------------------
	//make noise texture:
	(function(){
		noiseTexture = gl.createTexture();
		gl.bindTexture(gl.TEXTURE_2D, noiseTexture);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.REPEAT);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.REPEAT);
		var noise = new Uint8Array(128 * 128 * 3);
		for (var i = 0; i < 128 * 128 * 3; ++i) {
			noise[i] = Math.random() * 256;
		}
		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGB, 128, 128, 0, gl.RGB, gl.UNSIGNED_BYTE, noise);
		gl.bindTexture(gl.TEXTURE_2D, null);
	})();

	//----------------------------------------
	//make surface (noise) texture:
	(function(){
		surfaceTexture = gl.createTexture();
		gl.bindTexture(gl.TEXTURE_2D, surfaceTexture);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.REPEAT);
		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.REPEAT);
		var noise = new Uint8Array(1024 * 3);
		for (var i = 0; i + 2 < 1024 * 3; i += 3) {
			var ang1 = i / 1024 * Math.PI * 2.0 * 3.0;
			var ang2 = i / 1024 * Math.PI * 2.0 * 5.0;
			noise[i] = Math.cos(ang1) * 50 + Math.cos(ang2) * 30 + 128;
			noise[i+1] = Math.sin(ang1) * 50 + Math.sin(ang2) * 30 + 128;
			noise[i+2] = 0;
		}
		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGB, 1024, 1, 0, gl.RGB, gl.UNSIGNED_BYTE, noise);
		gl.bindTexture(gl.TEXTURE_2D, null);
	})();

	//----------------------------------------
	//load title texture:
	(function(){
		var i = new Image();
		i.addEventListener('load', function(){
			titleTexture = gl.createTexture();
			titleTexture.width = i.width;
			titleTexture.height = i.height;
			gl.bindTexture(gl.TEXTURE_2D, titleTexture);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
			gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, i);
			gl.bindTexture(gl.TEXTURE_2D, null);

		});
		i.src = titleTexData;
	})();

	(function(){
		var i = new Image();
		i.addEventListener('load', function(){
			wellDoneTexture = gl.createTexture();
			wellDoneTexture.width = i.width;
			wellDoneTexture.height = i.height;
			gl.bindTexture(gl.TEXTURE_2D, wellDoneTexture);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
			gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, i);
			gl.bindTexture(gl.TEXTURE_2D, null);

		});
		i.src = wellDoneTexData;
	})();



	//----------------------------------------
	//GL setup:

	gl.disable(gl.DEPTH_TEST);
	gl.clearColor(0.0, 0.0, 1.0, 1.0);
	gl.clear(gl.COLOR_BUFFER_BIT);

	//---------------------------------------
	//load scores:
	read_scores();

	//---------------------------------------
	//Init:
	set_select();

	//---------------------------------------

	window.addEventListener('mousedown', function(evt) {
		mousedown();

		evt.preventDefault();
		return false;
	});

	window.addEventListener('mouseup', function(evt) {
		mouseup();
		
		evt.preventDefault();
		return false;
	});

	window.addEventListener('mousemove', function(evt) {
		var rect = canvas.getBoundingClientRect();
		var x = evt.clientX - rect.left;
		var y = evt.clientY - rect.top;
		x = 2.0 * (x / canvas.width - 0.5);
		y = 2.0 * (0.5 - y / canvas.height);
		var pos = {
			x: camera.at.x + camera.right.x * x + camera.up.x * y,
			y: camera.at.y + camera.right.y * x + camera.up.y * y
		};
		mousemove(pos);

		evt.preventDefault();
		return false;
	});

	window.requestAnimFrame =
		window.requestAnimationFrame
		|| window.webkitRequestAnimationFrame
		|| window.mozRequestAnimationFrame
		|| window.oRequestAnimationFrame
		|| window.msRequestAnimationFrame
		|| function(callback) {
          window.setTimeout(callback, 1000 / 60);
        };
	window.requestAnimFrame(loop);
}
